import 'dart:math' as math;

import 'package:knt_bloc/knt_bloc.dart';

import '../model/book.dart';
import '../repo/book_repo.dart';

class BookBloc extends BasePagingBloc<void, Book> with Frontend<String> {
  final _books = <Book>[];
  bool _usingIsolator = false;

  BookBloc() {
    on<LoadBooksEvent>(handleLoadBooks);
  }

  @override
  bool get needLog => true;

  @override
  Future<BaseResponsePaging<Book>> pagingInProgress(
      BaseRequestPaging<void> pagingQuery) async {
    List<Book> partialBooks = _books.sublist(
      pagingQuery.offset,
      math.min(pagingQuery.offset + pagingQuery.limit, _books.length),
    );
    return BaseResponsePaging<Book>(dataList: partialBooks);
  }

  @override
  BasePagingResultState<Book> pagingStateInstance(
    List<Book> dataList,
    int offset, {
    String? tag,
  }) =>
      ShowBooksState(dataList, offset, tag: tag);

  @override
  Map<String, Function> get tasks => {
        (LoadBooksEvent).toString(): _setBooksFromBackend,
      };

  Future<void> handleLoadBooks(
    LoadBooksEvent event,
    Emitter<BaseState> emit,
  ) async {
    _usingIsolator = event.usingIsolate;
    emit(const BaseLoadingState());
    await Future.delayed(const Duration(milliseconds: 200));
    // throw const KntException(KntIssue.hostNotFound);
    if (_usingIsolator) {
      await initBackend(_backendFabric, backendType: BookBlocBackend);
      send(event.tag);
      return;
    }
    final loadedBook = await BookRepo().fetchBooks();
    _setBooksFromBackend(loadedBook);
  }

  void _setBooksFromBackend(List<Book> bookChunks) {
    _books.clear();
    _books.addAll(bookChunks);
    if (_usingIsolator) {
      killBackend();
    }
    add(const BasePagingEvent<void>(isFirstRequest: true));
  }
}

class BookBlocBackend extends Backend<String> {
  final _repo = BookRepo();

  BookBlocBackend(BackendArgument<void> argument) : super(argument);

  @override
  Map<String, Function> get operations => {
        (LoadBooksEvent).toString(): _loadBooks,
      };

  Future<void> _loadBooks() async {
    final books = await _repo.fetchBooks();
    await sendChunks((LoadBooksEvent).toString(), books);
  }
}

void _backendFabric(BackendArgument<void> argument) {
  BookBlocBackend(argument);
}

// region EVENT
class LoadBooksEvent extends BaseEvent {
  final bool usingIsolate;

  const LoadBooksEvent({
    this.usingIsolate = false,
  });
}
// endregion

// region STATE
class ShowBooksState extends BasePagingResultState<Book> {
  const ShowBooksState(
    List<Book> dataList,
    int currentOffset, {
    String? tag,
  }) : super(dataList, currentOffset, tag: tag);
}
// endregion
